
-- Create AddNetworkDeviceSysNameFirst
PRINT 'Creating procedure to add or update a network discovery device'
IF EXISTS(SELECT * FROM sysobjects WHERE xtype = 'P'
	AND name = 'AddNetworkDeviceSysNameFirst'
)
	DROP PROCEDURE AddNetworkDeviceSysNameFirst
GO

CREATE PROCEDURE AddNetworkDeviceSysNameFirst
	@AllowDuplicates BIT = 1,
	@ComputerID INT = NULL,
	@DeviceName NVARCHAR(64) = NULL,
	@DNSFullName NVARCHAR(256) = NULL,
	@NTDomainName NVARCHAR(256) = NULL,
	@IPAddress VARCHAR(64) = NULL,
	@IPSubnet VARCHAR(64) = NULL,
	@IPSubnetMask VARCHAR(64) = NULL,
	@PhysicalAddress VARCHAR(64) = NULL,
	@DeviceRole INT = NULL,
	@OperatingSystem NVARCHAR(128) = NULL,
	@OperatingSystemID INT = NULL,
	@IsManaged INT = NULL,
	@Description NVARCHAR(256) = NULL,
	@SystemDescription NVARCHAR(256) = NULL,
	@SystemLocation NVARCHAR(256) = NULL,
	@SystemContact NVARCHAR(256) = NULL,
	@LastDataSourceName NVARCHAR(128) = NULL,
	@LastDataSourceType NVARCHAR(32) = NULL,
	@DiscoveringServerUID BINARY(16) = NULL,
	@DiscoveringServerCanAdminister INT = 2, --unknown

	@OpenPortsTCP		varchar(512) = NULL,
	@OpenPortsUDP		varchar(512) = NULL,
	@ScannedOperatingSystem	nvarchar(512) = NULL,

	@ScannedOsType		nvarchar(512) = NULL,
	@ScannedOsVendor	nvarchar(512) = NULL,
	@ScannedOsFamily	nvarchar(512) = NULL,
	@ScannedOsGen		nvarchar(512) = NULL,
	@ScannedMacAddress	varchar(64) = NULL,
	@ScannedMacVendor	nvarchar(512) = NULL,

	@SNMP_snmpCommunityString nvarchar(256) = NULL,
	@SNMP_sysDescr		nvarchar(256) = NULL,
	@SNMP_sysObjectID	nvarchar(256) = NULL,
	@SNMP_sysObjectIDSymbolic nvarchar(256) = NULL,
	@SNMP_sysUpTime		bigint = NULL,
	@SNMP_sysContact	nvarchar(256) = NULL,
	@SNMP_sysName		nvarchar(256) = NULL,
	@SNMP_sysLocation	nvarchar(256) = NULL,
	@SNMP_sysServices	int = NULL,
	@SNMP_ipForwarding	int = NULL,

	@AccountID INT = NULL,

	@AMT_Secure INT = NULL,
	@AMT_Description NVARCHAR(256) = NULL,

	@Virtualization_Protocol VARCHAR(16) = NULL,
	@Virtualization_Port INT = NULL,
	@Virtualization_APIType VARCHAR(32) = NULL,
	@Virtualization_APIVersion VARCHAR(16) = NULL,
	@Virtualization_NameID INT = NULL,
	@Virtualization_Version VARCHAR(16) = NULL,
	@IsSQL INT = NULL,
	@SQLDiscoveredBy NVARCHAR(128) = NULL,
	@SQLPorts NVARCHAR(128) = NULL,

	@DeviceRoleOverride INT = NULL,
	@NoOidMatch INT = NULL,

	@DeviceID_OUT INT = NULL OUTPUT
AS
BEGIN
	PRINT 'D00 - Start AddNetworkDeviceSysNameFirst'
	-- We need at least one of DeviceName, DNSFullName, SNMP_sysName or PhysicalAddress
	PRINT 'D01 - Check if enougth Device information for Mapping'
	IF (@SNMP_sysName IS NULL 
		AND @PhysicalAddress IS NULL
		AND (@IPAddress IS NULL
			OR EXISTS (
			SELECT * FROM NetworkDevice AS nd
			LEFT OUTER JOIN NetworkDeviceSNMPInfo AS ndsi
			ON ndsi.DeviceID = nd.DeviceID
			WHERE IPAddress = @IPAddress 
			AND (
			ndsi.SNMP_sysName IS NOT NULL 
			OR PhysicalAddress IS NOT NULL
			)))
		)
		BEGIN
			PRINT 'E01 - DeviceName, DNSFullName, SNMP_sysName, PhysicalAddress or IPAddress is null'
			PRINT 'E01 - or IP-Address exists with more Information as only IP-Address'
			RETURN
		END

	-- Make sure we overwrite IP-only discovery records with more complete data		
	PRINT 'D02 - Make sure w overwrite IP-only discovery records with more complete data'
	IF (@AllowDuplicates = 1
		AND EXISTS (
			SELECT * FROM NetworkDevice AS nd
			LEFT OUTER JOIN NetworkDeviceSNMPInfo AS ndsi
			ON ndsi.DeviceID = nd.DeviceID
			WHERE IPAddress = @IPAddress 
			AND DeviceName IS NULL 
			AND DNSFullName IS NULL 
			AND ndsi.SNMP_sysName IS NULL 
			AND PhysicalAddress IS NULL))
		SET @AllowDuplicates = 0

	PRINT 'D02 - AddNetworkDevice DeviceName = '
		+ ISNULL(@DeviceName, '<null>')
		+ ' / IPAddress = ' + ISNULL(@IPAddress, '<null>')
		+ ' / PhysicalAddress = ' + ISNULL(@PhysicalAddress, '<null>')
		+ ' / DNSFullName = ' + ISNULL(@DNSFullName, '<null>')
		+ ' / SNMP_sysName = ' + ISNULL(@SNMP_sysName, '<null>') 
		+ ' / NTDomainName = ' + ISNULL(@NTDomainName, '<null>') 
		+ ' / AllowDuplicates = ' + CAST(@AllowDuplicates AS NVARCHAR(1))

	-- Populate the operating system name and ID: Trust the ID more!
	PRINT 'D03 - Populate the operating system name and ID: ID has higher whight'
	IF (@OperatingSystemID IS NOT NULL)
		SELECT @OperatingSystem = OperatingSystemName
			FROM OperatingSystem
			WHERE OperatingSystemID = @OperatingSystemID
	ELSE IF (@OperatingSystem IS NOT NULL)
		SELECT @OperatingSystemID = OperatingSystemID
			FROM OperatingSystem
			WHERE OperatingSystemName = @OperatingSystem
	PRINT 'D03 - OperatingSystemID = ' + ISNULL(CAST(@OperatingSystemID AS NVARCHAR(16)), '<null>') + ' / OperatingSystem = ' + ISNUll(@OperatingSystem, '<null>')

	-- It is possible to have separate rows representing the same
	-- device, for example, if an IP and no name was added and a
	-- name with no IP was added.  If later this procedure gets
	-- called with both name and IP then we have to merge the rows
	-- together.  We do this in order of LastUpdate so that the
	-- merged row is the most recent information.  Note that
	-- because we have 3 keys used to lookup a device we may have
	-- to merge as many as 3 rows into 1.

	PRINT 'D04 - Start Merging Process'
	-- Variables for merge
	DECLARE @PrimaryDeviceID	INT
	DECLARE @FirstDeviceID		INT
	DECLARE @FirstDeviceDuplicateID INT
	DECLARE @OldDeviceID		INT
	DECLARE @OldDuplicateID		INT
	DECLARE @OldComputerID		INT
	DECLARE @OldDeviceName		NVARCHAR(64)
	DECLARE @OldDNSFullName		NVARCHAR(256)
	DECLARE @OldNTDomainName	NVARCHAR(256)
	DECLARE @OldIPAddress		VARCHAR(64)
	DECLARE @OldIPSubnet		VARCHAR(64)
	DECLARE @OldIPSubnetMask	VARCHAR(64)
	DECLARE @OldPhysicalAddress	VARCHAR(64)
	DECLARE @OldDeviceRole		INT
	DECLARE @OldOperatingSystem	NVARCHAR(128)
	DECLARE @OldOperatingSystemID	INT
	DECLARE @OldIsManaged		INT
	DECLARE @OldDescription		NVARCHAR(256)
	DECLARE @OldSystemDescription	NVARCHAR(256)
	DECLARE @OldSystemLocation	NVARCHAR(256)
	DECLARE @OldSystemContact	NVARCHAR(256)
	DECLARE @OldFirstDiscovered	DATETIME
	DECLARE @OldLastUpdate		DATETIME
	DECLARE @OldLastDataSourceName	NVARCHAR(128)
	DECLARE @OldLastDataSourceType	NVARCHAR(32)
	DECLARE @FirstDiscovered	DATETIME
	DECLARE @InvOperatingSystemID	INT
	DECLARE @IsBetter		BIT	-- Result from OperatingSystem category check

	DECLARE @OldOpenPortsTCP			varchar(512)
	DECLARE @OldOpenPortsUDP			varchar(512)
	DECLARE @OldScannedOperatingSystem	nvarchar(512)

	DECLARE @OldScannedOsType		nvarchar(512)
	DECLARE @OldScannedOsVendor		nvarchar(512)
	DECLARE @OldScannedOsFamily		nvarchar(512)
	DECLARE @OldScannedOsGen		nvarchar(512)
	DECLARE @OldScannedMacAddress	varchar(64)
	DECLARE @OldScannedMacVendor	nvarchar(512)

	DECLARE @OldSNMP_snmpCommunityString nvarchar(256)
	DECLARE @OldSNMP_sysDescr			nvarchar(256)
	DECLARE @OldSNMP_sysObjectID		nvarchar(256)
	DECLARE @OldSNMP_sysObjectIDSymbolic nvarchar(256)
	DECLARE @OldSNMP_sysUpTime			bigint
	DECLARE @OldSNMP_sysContact			nvarchar(256)
	DECLARE @OldSNMP_sysName			nvarchar(256)
	DECLARE @OldSNMP_sysLocation		nvarchar(256)
	DECLARE @OldSNMP_sysServices		int
	DECLARE @OldSNMP_ipForwarding		int

	DECLARE @OldAMT_Secure		int
	DECLARE @OldAMT_Description		nvarchar(256)

	DECLARE @OldVirtualization_Protocol VARCHAR(16)
	DECLARE @OldVirtualization_Port INT
	DECLARE @OldVirtualization_APIType VARCHAR(32)
	DECLARE @OldVirtualization_APIVersion VARCHAR(16)
	DECLARE @OldVirtualization_NameID INT
	DECLARE @OldVirtualization_Version VARCHAR(16)
	
	DECLARE @OldDeviceRoleOverride INT
	DECLARE @OldNoOidMatch INT

	-- Begin transaction
	BEGIN TRANSACTION

	-- Open cursor for merging duplicates
	IF @AllowDuplicates = 1
	BEGIN
		PRINT 'D05 - Get Dupplicated Devices (AllowDuplicates = 1)'
		-- Look for duplicate primary device
		SELECT	@PrimaryDeviceID = nd.DeviceID
		FROM	NetworkDevice nd
		LEFT OUTER JOIN	NetworkDeviceSNMPInfo AS ndsi
			ON ndsi.DeviceID = nd.DeviceID
		WHERE	DuplicateID IS NULL
			AND (
				-- matching on PhysicalAddress REGARDLESS
				PhysicalAddress = @PhysicalAddress
				
				-- matching on SNMP_sysName REGARDLESS
				OR SNMP_sysName = @SNMP_sysName
			)
		PRINT 'D05 - Get PrimaryDeviceID: ' + @PrimaryDeviceID
		PRINT 'D05 - Create DuplicateCursor with exact match on all identifying values'

		-- Exact match on all identifying values
		DECLARE DuplicateCursor CURSOR FOR
		SELECT	nd.DeviceID,
			nd.DuplicateID,
			nd.ComputerID,
			nd.DeviceName,
			nd.DNSFullName,
			nd.NTDomainName,
			nd.IPAddress,
			nd.IPSubnet,
			nd.IPSubnetMask,
			nd.PhysicalAddress,
			nd.DeviceRole,
			nd.OperatingSystem,
			nd.OperatingSystemID,
			nd.IsManaged,
			nd.Description,
			nd.SystemDescription,
			nd.SystemLocation,
			nd.SystemContact,
			nd.FirstDiscovered,
			nd.LastUpdate,
			nd.LastDataSourceName,
			nd.LastDataSourceType,

			nd.OpenPortsTCP,
			nd.OpenPortsUDP,
			nd.ScannedOperatingSystem,

			nd.ScannedOsType,
			nd.ScannedOsVendor,
			nd.ScannedOsFamily,
			nd.ScannedOsGen,
			nd.ScannedMacAddress,
			nd.ScannedMacVendor,
			nd.DeviceRoleOverride,
			nd.NoOidMatch,

			ndsi.SNMP_snmpCommunityString,
			ndsi.SNMP_sysDescr,
			ndsi.SNMP_sysObjectID,
			ndsi.SNMP_sysObjectIDSymbolic,
			ndsi.SNMP_sysUpTime,
			ndsi.SNMP_sysContact,
			ndsi.SNMP_sysName,
			ndsi.SNMP_sysLocation,
			ndsi.SNMP_sysServices,
			ndsi.SNMP_ipForwarding,

			ndai.Secure AS AMT_Secure,
			ndai.Description AS AMT_Description,

			ndvi.Protocol AS Virtualization_Protocol,
			ndvi.Port AS Virtualization_Port,
			ndvi.APIType AS Virtualization_APIType,
			ndvi.APIVersion AS Virtualization_APIVersion,
			ndvi.ProductNameID AS Virtualization_NameID,
			ndvi.ProductVersion AS Virtualization_Version

		FROM	NetworkDevice AS nd
		LEFT OUTER JOIN	NetworkDeviceSNMPInfo AS ndsi
			ON ndsi.DeviceID = nd.DeviceID
		LEFT OUTER JOIN	NetworkDeviceAMTInfo AS ndai
			ON ndai.DeviceID = nd.DeviceID
		LEFT OUTER JOIN	NetworkDeviceVirtualizationInfo AS ndvi
			ON ndvi.DeviceID = nd.DeviceID
		WHERE	nd.PhysicalAddress = @PhysicalAddress OR ndsi.SNMP_sysName = @SNMP_sysName
			
		ORDER BY nd.LastUpdate DESC
	END
	ELSE
	BEGIN
		-- PrimarayDeviceID is only needed if for one Device exist more then one NetworkDevice Entry (AllowDuplicates = 1)
		PRINT 'D05 - Get Dupplicated Devices (AllowDuplicates = 0)'
		-- Match up on at least one identifying value
		DECLARE DuplicateCursor CURSOR FOR
		SELECT	nd.DeviceID,
			nd.DuplicateID,
			nd.ComputerID,
			nd.DeviceName,
			nd.DNSFullName,
			nd.NTDomainName,
			nd.IPAddress,
			nd.IPSubnet,
			nd.IPSubnetMask,
			nd.PhysicalAddress,
			nd.DeviceRole,
			nd.OperatingSystem,
			nd.OperatingSystemID,
			nd.IsManaged,
			nd.Description,
			nd.SystemDescription,
			nd.SystemLocation,
			nd.SystemContact,
			nd.FirstDiscovered,
			nd.LastUpdate,
			nd.LastDataSourceName,
			nd.LastDataSourceType,

			nd.OpenPortsTCP,
			nd.OpenPortsUDP,
			nd.ScannedOperatingSystem,

			nd.ScannedOsType,
			nd.ScannedOsVendor,
			nd.ScannedOsFamily,
			nd.ScannedOsGen,
			nd.ScannedMacAddress,
			nd.ScannedMacVendor,
			nd.DeviceRoleOverride,
			nd.NoOidMatch,

			ndsi.SNMP_snmpCommunityString,
			ndsi.SNMP_sysDescr,
			ndsi.SNMP_sysObjectID,
			ndsi.SNMP_sysObjectIDSymbolic,
			ndsi.SNMP_sysUpTime,
			ndsi.SNMP_sysContact,
			ndsi.SNMP_sysName,
			ndsi.SNMP_sysLocation,
			ndsi.SNMP_sysServices,
			ndsi.SNMP_ipForwarding,

			ndai.Secure AS AMT_Secure,
			ndai.Description AS AMT_Description,

			ndvi.Protocol AS Virtualization_Protocol,
			ndvi.Port AS Virtualization_Port,
			ndvi.APIType AS Virtualization_APIType,
			ndvi.APIVersion AS Virtualization_APIVersion,
			ndvi.ProductNameID AS Virtualization_NameID,
			ndvi.ProductVersion AS Virtualization_Version

		FROM	NetworkDevice AS nd
		LEFT OUTER JOIN	NetworkDeviceSNMPInfo AS ndsi
			ON ndsi.DeviceID = nd.DeviceID
		LEFT OUTER JOIN	NetworkDeviceAMTInfo AS ndai
			ON ndai.DeviceID = nd.DeviceID
		LEFT OUTER JOIN	NetworkDeviceVirtualizationInfo AS ndvi
			ON ndvi.DeviceID = nd.DeviceID

		-- Note: should all this logic be changed, please make sure to update NetworkDeviceFindDuplicateToMergeWith further down
		INNER JOIN (
			-- matching on PhysicalAddress REGARDLESS
			SELECT DeviceID
			FROM NetworkDevice nd2
			WHERE nd2.DuplicateID IS NULL
				AND nd2.PhysicalAddress = @PhysicalAddress

			UNION

			-- matching on SNMP_sysName, if there are no contradictions
			SELECT nd2.DeviceID
			FROM NetworkDevice nd2
			LEFT OUTER JOIN NetworkDeviceSNMPInfo AS ndsi2
			ON ndsi2.DeviceID = nd2.DeviceID
			WHERE nd2.DuplicateID IS NULL
					AND ndsi2.SNMP_sysName = @SNMP_sysName

		) as matches
			ON matches.DeviceID = nd.DeviceID

		ORDER BY nd.LastUpdate DESC
	END

	PRINT 'D06 - Work on DuplicateCursor'
	OPEN DuplicateCursor
	FETCH NEXT FROM DuplicateCursor
	INTO	@OldDeviceID,
		@OldDuplicateID,
		@OldComputerID,
		@OldDeviceName,
		@OldDNSFullName,
		@OldNTDomainName,
		@OldIPAddress,
		@OldIPSubnet,
		@OldIPSubnetMask,
		@OldPhysicalAddress,
		@OldDeviceRole,
		@OldOperatingSystem,
		@OldOperatingSystemID,
		@OldIsManaged,
		@OldDescription,
		@OldSystemDescription,
		@OldSystemLocation,
		@OldSystemContact,
		@OldFirstDiscovered,
		@OldLastUpdate,
		@OldLastDataSourceName,
		@OldLastDataSourceType,

		@OldOpenPortsTCP,
		@OldOpenPortsUDP,
		@OldScannedOperatingSystem,

		@OldScannedOsType,
		@OldScannedOsVendor,
		@OldScannedOsFamily,
		@OldScannedOsGen,
		@OldScannedMacAddress,
		@OldScannedMacVendor,
		@OldDeviceRoleOverride,
		@OldNoOidMatch,

		@OldSNMP_snmpCommunityString,
		@OldSNMP_sysDescr,
		@OldSNMP_sysObjectID,
		@OldSNMP_sysObjectIDSymbolic,
		@OldSNMP_sysUpTime,
		@OldSNMP_sysContact,
		@OldSNMP_sysName,
		@OldSNMP_sysLocation,
		@OldSNMP_sysServices,
		@OldSNMP_ipForwarding,

		@OldAMT_Secure,
		@OldAMT_Description,

		@OldVirtualization_Protocol,
		@OldVirtualization_Port,
		@OldVirtualization_APIType,
		@OldVirtualization_APIVersion,
		@OldVirtualization_NameID,
		@OldVirtualization_Version

		PRINT 'D06 - Merge all Information from all founded devices, use the best information'
		PRINT 'D06 - FirstDeviceID = ' + ISNULL(CAST(@FirstDeviceID AS NVARCHAR(16)), '<null>') +
				' / ComputerID = ' + ISNULL(CAST(@ComputerID AS NVARCHAR(16)), '<null>') +
				' / DeviceName = ' + ISNULL(@DeviceName, '<null>') +
				' / DNSFullName = ' + ISNULL(@DNSFullName, '<null>') +
				' / NTDomainName = ' + ISNULL(@NTDomainName, '<null>') +
				' / IPAddress = ' + ISNULL(@IPAddress, '<null>') +
				' / IPSubnet = ' + ISNULL(@IPSubnet, '<null>') +
				' / IPSubnetMask = ' + ISNULL(@IPSubnetMask, '<null>') +
				' / PhysicalAddress = ' + ISNULL(@PhysicalAddress, '<null>') +
				' / DeviceRole = ' + ISNULL(CAST(@DeviceRole AS NVARCHAR(16)), '<null>') +
				' / NoOidMatch = ' + ISNULL(CAST(@NoOidMatch AS NVARCHAR(16)), '<null>') +
				' / OperatingSystemID = ' + ISNULL(CAST(@OperatingSystemID AS NVARCHAR(16)), '<null>') +
				' / OperatingSystem = ' + ISNULL(@OperatingSystem, '<null>') +
				' / IsManaged = ' + ISNULL(CAST(@IsManaged AS NVARCHAR(16)), '<null>') +
				' / Description = ' + ISNULL(@Description, '<null>') +
				' / SystemDescription = ' + ISNULL(@SystemDescription, '<null>') +
				' / SystemLocation = ' + ISNULL(@SystemLocation, '<null>') +
				' / SystemContact = ' + ISNULL(@SystemContact, '<null>') +
				' / FirstDiscovered = ' + ISNULL(CAST(@FirstDiscovered AS NVARCHAR(32)), '<null>') +
				' / LastDataSourceName = ' + ISNULL(@LastDataSourceName, '<null>') +
				' / LastDataSourceType = ' + ISNULL(CAST(@LastDataSourceType AS NVARCHAR(16)), '<null>') +
				' / OpenPortsTCP = ' + ISNULL(@OpenPortsTCP, '<null>') +
				' / OpenPortsUDP = ' + ISNULL(@OpenPortsUDP, '<null>') +
				' / ScannedOperatingSystem = ' + ISNULL(@ScannedOperatingSystem, '<null>') +
				' / ScannedOSType = ' + ISNULL(@ScannedOSType, '<null>') +
				' / ScannedOSVendor = ' + ISNULL(@ScannedOSVendor, '<null>') +
				' / ScannedOSFamily = ' + ISNULL(@ScannedOSFamily, '<null>') +
				' / ScannedOSGen = ' + ISNULL(@ScannedOSGen, '<null>') +
				' / ScannedMacAddress = ' + ISNULL(@ScannedMacAddress, '<null>') +
				' / ScannedMacVendor = ' + ISNULL(@ScannedMacVendor, '<null>') +
				' / SNMP_snmpCommunityString = ' + ISNULL(@SNMP_snmpCommunityString, '<null>') +
				' / SNMP_sysDescr = ' + ISNULL(@SNMP_sysDescr, '<null>') +
				' / SNMP_sysObjectID = ' + ISNULL(@SNMP_sysObjectID, '<null>') +
				' / SNMP_sysObjectIDSymbolic = ' + ISNULL(@SNMP_sysObjectIDSymbolic, '<null>') +
				' / SNMP_sysUpTime = ' + ISNULL(CAST(@SNMP_sysUpTime AS NVARCHAR(32)), '<null>') +
				' / SNMP_sysContact = ' + ISNULL(@SNMP_sysContact, '<null>') +
				' / SNMP_sysName = ' + ISNULL(@SNMP_sysName, '<null>') +
				' / SNMP_sysLocation = ' + ISNULL(@SNMP_sysLocation, '<null>') +
				' / SNMP_sysServices = ' + ISNULL(CAST(@SNMP_sysServices AS NVARCHAR(16)), '<null>') +
				' / SNMP_ipForwarding = ' + ISNULL(CAST(@SNMP_ipForwarding AS NVARCHAR(16)), '<null>') +
				' / AMT_Secure = ' + ISNULL(CAST(@AMT_Secure AS NVARCHAR(16)), '<null>') +
				' / AMT_Description = ' + ISNULL(@AMT_Description, '<null>') +
				' / Virtualization_Protocol = ' + ISNULL(@Virtualization_Protocol, '<null>') +
				' / Virtualization_Port = ' + ISNULL(CAST(@Virtualization_Port AS NVARCHAR(16)), '<null>') +
				' / Virtualization_APIType = ' + ISNULL(@Virtualization_APIType, '<null>') +
				' / Virtualization_APIVersion = ' + ISNULL(@Virtualization_APIVersion, '<null>') +
				' / Virtualization_NameID = ' + ISNULL(CAST(@Virtualization_NameID AS NVARCHAR(16)), '<null>') +
				' / Virtualization_Version = ' + ISNULL(@Virtualization_Version, '<null>') +
				' / DeviceRoleOverride = ' + ISNULL(CAST(@DeviceRoleOverride AS NVARCHAR(16)), '<null>') + ' (StartDataSet)'

	IF @@FETCH_STATUS = 0
	BEGIN
		SET @FirstDeviceID = @OldDeviceID
		SET @FirstDeviceDuplicateID = @OldDuplicateID
		SET @FirstDiscovered = GETDATE()

		-- Merge information from all rows that look like duplicates
		WHILE 1 = 1
		BEGIN
			-- Overwrite NULLs with information from matching rows
			-- Use for ComputerID allways the oldest ComputerID
			IF @ComputerID IS NULL
				BEGIN
					IF @OldComputerID IS NOT NULL SET @ComputerID = @OldComputerID
				END
			ELSE
				BEGIN
					IF @ComputerID > @OldComputerID SET @ComputerID = @OldComputerID
				END
			
			SET @DeviceName = ISNULL(
				@DeviceName, @OldDeviceName
			)
			SET @DNSFullName = ISNULL(
				@DNSFullName, @OldDNSFullName
			)
			SET @NTDomainName = ISNULL(
				@NTDomainName, @OldNTDomainName
			)
			SET @IPAddress = ISNULL(
				@IPAddress, @OldIPAddress
			)
			SET @IPSubnet = ISNULL(
				@IPSubnet, @OldIPSubnet
			)
			SET @IPSubnetMask = ISNULL(
				@IPSubnetMask, @OldIPSubnetMask
			)
			SET @PhysicalAddress = ISNULL(
				@PhysicalAddress, @OldPhysicalAddress
			)
			SET @DeviceRole = CASE WHEN @OldDeviceRoleOverride = 1 THEN @OldDeviceRole WHEN @OldDeviceRole > 0 AND @DeviceRole = 0 THEN @OldDeviceRole ELSE ISNULL(@DeviceRole, @OldDeviceRole) END
			SET @NoOidMatch = ISNULL(
				@NoOidMatch, @OldNoOidMatch
			)

			-- If OldOperatingSystemID is better/more refined, use it instead
			SET @IsBetter = CASE WHEN (@OperatingSystemID IS NULL
						OR  @OperatingSystemID <= 1)
					THEN 1	-- Anything's better than NULL/unknown/other
					ELSE 0
					END
			IF @IsBetter = 0	--  Old may be better:
				-- If OldOperatingSystem looks ok, is it really better?
				EXEC @IsBetter = OperatingSystemCategoryTest
							@OperatingSystemID,
							@OldOperatingSystemID
			IF @IsBetter <> 0	-- Old definitely *is* better
			BEGIN
				SET @OperatingSystemID = @OldOperatingSystemID
				SELECT @OperatingSystem = OperatingSystemName
					FROM OperatingSystem
					WHERE OperatingSystemID = @OperatingSystemID
			END

			SET @IsManaged = ISNULL(
				@IsManaged, @OldIsManaged
			)
			SET @Description = ISNULL(
				@Description, @OldDescription
			)
			SET @SystemDescription = ISNULL(
				@SystemDescription, @OldSystemDescription
			)
			SET @SystemLocation = ISNULL(
				@SystemLocation, @OldSystemLocation
			)
			SET @SystemContact = ISNULL(
				@SystemContact, @OldSystemContact
			)
			IF (@OldFirstDiscovered < @FirstDiscovered)
				SET @FirstDiscovered = @OldFirstDiscovered
			SET @LastDataSourceName = ISNULL(
				@LastDataSourceName, @OldLastDataSourceName
			)
			SET @LastDataSourceType = ISNULL(
				@LastDataSourceType, @OldLastDataSourceType
			)

			SET @OpenPortsTCP = ISNULL(
				@OpenPortsTCP, @OldOpenPortsTCP
			)
			SET @OpenPortsUDP = ISNULL(
				@OpenPortsUDP, @OldOpenPortsUDP
			)
			SET @ScannedOperatingSystem = ISNULL(
				@ScannedOperatingSystem, @OldScannedOperatingSystem
			)

			SET @ScannedOsType = ISNULL(
				@ScannedOsType, @OldScannedOsType
			)
			SET @ScannedOsVendor = ISNULL(
				@ScannedOsVendor, @OldScannedOsVendor
			)
			SET @ScannedOsFamily = ISNULL(
				@ScannedOsFamily, @OldScannedOsFamily
			)
			SET @ScannedOsGen = ISNULL(
				@ScannedOsGen, @OldScannedOsGen
			)
			SET @ScannedMacAddress = ISNULL(
				@ScannedMacAddress, @OldScannedMacAddress
			)
			SET @ScannedMacVendor = ISNULL(
				@ScannedMacVendor, @OldScannedMacVendor
			)

			SET @SNMP_snmpCommunityString = ISNULL(
				@SNMP_snmpCommunityString, @OldSNMP_snmpCommunityString
			)
			SET @SNMP_sysDescr = ISNULL(
				@SNMP_sysDescr, @OldSNMP_sysDescr
			)
			SET @SNMP_sysObjectID = ISNULL(
				@SNMP_sysObjectID, @OldSNMP_sysObjectID
			)
			SET @SNMP_sysObjectIDSymbolic = ISNULL(
				@SNMP_sysObjectIDSymbolic, @OldSNMP_sysObjectIDSymbolic
			)
			SET @SNMP_sysUpTime = ISNULL(
				@SNMP_sysUpTime, @OldSNMP_sysUpTime
			)
			SET @SNMP_sysContact = ISNULL(
				@SNMP_sysContact, @OldSNMP_sysContact
			)
			SET @SNMP_sysName = ISNULL(
				@SNMP_sysName, @OldSNMP_sysName
			)
			SET @SNMP_sysLocation = ISNULL(
				@SNMP_sysLocation, @OldSNMP_sysLocation
			)
			SET @SNMP_sysServices = ISNULL(
				@SNMP_sysServices, @OldSNMP_sysServices
			)
			SET @SNMP_ipForwarding = ISNULL(
				@SNMP_ipForwarding, @OldSNMP_ipForwarding
			)

			SET @AMT_Secure = ISNULL(
				@AMT_Secure, @OldAMT_Secure
			)
			SET @AMT_Description = ISNULL(
				@AMT_Description, @OldAMT_Description
			)

			SET @Virtualization_Protocol = ISNULL(
				@Virtualization_Protocol, @OldVirtualization_Protocol
			)
			SET @Virtualization_Port = ISNULL(
				@Virtualization_Port, @OldVirtualization_Port
			)
			SET @Virtualization_APIType = ISNULL(
				@Virtualization_APIType, @OldVirtualization_APIType
			)
			SET @Virtualization_APIVersion = ISNULL(
				@Virtualization_APIVersion, @OldVirtualization_APIVersion
			)
			SET @Virtualization_NameID = ISNULL(
				@Virtualization_NameID, @OldVirtualization_NameID
			)
			SET @Virtualization_Version = ISNULL(
				@Virtualization_Version, @OldVirtualization_Version
			)

			SET @DeviceRoleOverride = ISNULL(
				@DeviceRoleOverride, @OldDeviceRoleOverride
			)
			
			SET @NoOidMatch = ISNULL(
				@NoOidMatch, @OldNoOidMatch
			)

		PRINT 'D06 - OldDeviceID = ' + ISNULL(CAST(@OldDeviceID AS NVARCHAR(16)), '<null>') +
				' / ComputerID = ' + ISNULL(CAST(@ComputerID AS NVARCHAR(16)), '<null>')+
				' / DeviceName = ' + ISNULL(@DeviceName, '<null>') +
				' / DNSFullName = ' + ISNULL(@DNSFullName, '<null>') +
				' / NTDomainName = ' + ISNULL(@NTDomainName, '<null>') +
				' / IPAddress = ' + ISNULL(@IPAddress, '<null>') +
				' / IPSubnet = ' + ISNULL(@IPSubnet, '<null>') +
				' / IPSubnetMask = ' + ISNULL(@IPSubnetMask, '<null>') +
				' / PhysicalAddress = ' + ISNULL(@PhysicalAddress, '<null>') +
				' / DeviceRole = ' + ISNULL(CAST(@DeviceRole AS NVARCHAR(16)), '<null>') +
				' / NoOidMatch = ' + ISNULL(CAST(@NoOidMatch AS NVARCHAR(16)), '<null>') +
				' / OperatingSystemID = ' + ISNULL(CAST(@OperatingSystemID AS NVARCHAR(16)), '<null>') +
				' / OperatingSystem = ' + ISNULL(@OperatingSystem, '<null>') +
				' / IsManaged = ' + ISNULL(CAST(@IsManaged AS NVARCHAR(16)), '<null>') +
				' / Description = ' + ISNULL(@Description, '<null>') +
				' / SystemDescription = ' + ISNULL(@SystemDescription, '<null>') +
				' / SystemLocation = ' + ISNULL(@SystemLocation, '<null>') +
				' / SystemContact = ' + ISNULL(@SystemContact, '<null>') +
				' / FirstDiscovered = ' + ISNULL(CAST(@FirstDiscovered AS NVARCHAR(32)), '<null>') +
				' / LastDataSourceName = ' + ISNULL(@LastDataSourceName, '<null>') +
				' / LastDataSourceType = ' + ISNULL(CAST(@LastDataSourceType AS NVARCHAR(16)), '<null>') +
				' / OpenPortsTCP = ' + ISNULL(@OpenPortsTCP, '<null>') +
				' / OpenPortsUDP = ' + ISNULL(@OpenPortsUDP, '<null>') +
				' / ScannedOperatingSystem = ' + ISNULL(@ScannedOperatingSystem, '<null>') +
				' / ScannedOSType = ' + ISNULL(@ScannedOSType, '<null>') +
				' / ScannedOSVendor = ' + ISNULL(@ScannedOSVendor, '<null>') +
				' / ScannedOSFamily = ' + ISNULL(@ScannedOSFamily, '<null>') +
				' / ScannedOSGen = ' + ISNULL(@ScannedOSGen, '<null>') +
				' / ScannedMacAddress = ' + ISNULL(@ScannedMacAddress, '<null>') +
				' / ScannedMacVendor = ' + ISNULL(@ScannedMacVendor, '<null>') +
				' / SNMP_snmpCommunityString = ' + ISNULL(@SNMP_snmpCommunityString, '<null>') +
				' / SNMP_sysDescr = ' + ISNULL(@SNMP_sysDescr, '<null>') +
				' / SNMP_sysObjectID = ' + ISNULL(@SNMP_sysObjectID, '<null>') +
				' / SNMP_sysObjectIDSymbolic = ' + ISNULL(@SNMP_sysObjectIDSymbolic, '<null>') +
				' / SNMP_sysUpTime = ' + ISNULL(CAST(@SNMP_sysUpTime AS NVARCHAR(32)), '<null>') +
				' / SNMP_sysContact = ' + ISNULL(@SNMP_sysContact, '<null>') +
				' / SNMP_sysName = ' + ISNULL(@SNMP_sysName, '<null>') +
				' / SNMP_sysLocation = ' + ISNULL(@SNMP_sysLocation, '<null>') +
				' / SNMP_sysServices = ' + ISNULL(CAST(@SNMP_sysServices AS NVARCHAR(16)), '<null>') +
				' / SNMP_ipForwarding = ' + ISNULL(CAST(@SNMP_ipForwarding AS NVARCHAR(16)), '<null>') +
				' / AMT_Secure = ' + ISNULL(CAST(@AMT_Secure AS NVARCHAR(16)), '<null>') +
				' / AMT_Description = ' + ISNULL(@AMT_Description, '<null>') +
				' / Virtualization_Protocol = ' + ISNULL(@Virtualization_Protocol, '<null>') +
				' / Virtualization_Port = ' + ISNULL(CAST(@Virtualization_Port AS NVARCHAR(16)), '<null>') +
				' / Virtualization_APIType = ' + ISNULL(@Virtualization_APIType, '<null>') +
				' / Virtualization_APIVersion = ' + ISNULL(@Virtualization_APIVersion, '<null>') +
				' / Virtualization_NameID = ' + ISNULL(CAST(@Virtualization_NameID AS NVARCHAR(16)), '<null>') +
				' / Virtualization_Version = ' + ISNULL(@Virtualization_Version, '<null>') +
				' / DeviceRoleOverride = ' + ISNULL(CAST(@DeviceRoleOverride AS NVARCHAR(16)), '<null>')
			
			-- Fetch next row
			FETCH NEXT FROM DuplicateCursor
			INTO	@OldDeviceID,
				@OldDuplicateID,
				@OldComputerID,
				@OldDeviceName,
				@OldDNSFullName,
				@OldNTDomainName,
				@OldIPAddress,
				@OldIPSubnet,
				@OldIPSubnetMask,
				@OldPhysicalAddress,
				@OldDeviceRole,
				@OldOperatingSystem,
				@OldOperatingSystemID,
				@OldIsManaged,
				@OldDescription,
				@OldSystemDescription,
				@OldSystemLocation,
				@OldSystemContact,
				@OldFirstDiscovered,
				@OldLastUpdate,
				@OldLastDataSourceName,
				@OldLastDataSourceType,

				@OldOpenPortsTCP,
				@OldOpenPortsUDP,
				@OldScannedOperatingSystem,

				@OldScannedOsType,
				@OldScannedOsVendor,
				@OldScannedOsFamily,
				@OldScannedOsGen,
				@OldScannedMacAddress,
				@OldScannedMacVendor,
				@OldDeviceRoleOverride,
				@OldNoOidMatch,

				@OldSNMP_snmpCommunityString,
				@OldSNMP_sysDescr,
				@OldSNMP_sysObjectID,
				@OldSNMP_sysObjectIDSymbolic,
				@OldSNMP_sysUpTime,
				@OldSNMP_sysContact,
				@OldSNMP_sysName,
				@OldSNMP_sysLocation,
				@OldSNMP_sysServices,
				@OldSNMP_ipForwarding,

				@OldAMT_Secure,
				@OldAMT_Description,

				@OldVirtualization_Protocol,
				@OldVirtualization_Port,
				@OldVirtualization_APIType,
				@OldVirtualization_APIVersion,
				@OldVirtualization_NameID,
				@OldVirtualization_Version

			IF @@FETCH_STATUS <> 0
			BEGIN
				PRINT 'D07 - nothing to Remove'
				BREAK
			END

			-- Delete all but first row

			-- We are about to delete the primary device so
			-- make the first matching device the new
			-- primary device
			IF (@OldDeviceID = @PrimaryDeviceID)
				SET @PrimaryDeviceID = @FirstDeviceID

			PRINT 'D07 - Remove not more needed NetworkDiscovery Entries'
			PRINT 'D07 - Update First Device Set with DuplicateID = ' + CAST(@FirstDeviceID AS NVARCHAR(16)) + ' where DupplicateID = ' + CAST(@OldDeviceID AS NVARCHAR(16))
			-- Redirect any references stopping us from
			-- deleting row
			UPDATE	NetworkDevice
			SET	DuplicateID = @FirstDeviceID
			WHERE	DuplicateID = @OldDeviceID

			PRINT 'D07 - NetworkDeviceDiscoverdBy information for DeviceID = ' + CAST(@OldDeviceID AS NVARCHAR(16))
			-- Delete DiscoveredBy information for the device
			DELETE FROM NetworkDeviceDiscoveredBy
			WHERE NetworkDeviceDiscoveredBy.DeviceID = @OldDeviceID
			AND NOT EXISTS(
				SELECT ActionApplies.DeviceID
				FROM ActionApplies
				WHERE ActionApplies.DeviceID = @OldDeviceID
			)

			PRINT 'D07 - NetworkDeviceParent information for DeviceID = ' + CAST(@OldDeviceID AS NVARCHAR(16))
			DELETE FROM NetworkDeviceParent
			WHERE (DeviceID = @OldDeviceID OR ParentDeviceID = @OldDeviceID)
			AND NOT EXISTS(
				SELECT ActionApplies.DeviceID
				FROM ActionApplies
				WHERE ActionApplies.DeviceID = @OldDeviceID
			)
			
			-- Delete row
			PRINT 'D07 - NetworkDevice information for DeviceID = ' + CAST(@OldDeviceID AS NVARCHAR(16))
			DELETE FROM NetworkDevice
			WHERE NetworkDevice.DeviceID = @OldDeviceID
			AND NOT EXISTS(
				SELECT ActionApplies.DeviceID
				FROM ActionApplies
				WHERE ActionApplies.DeviceID = @OldDeviceID
			)
		END

		-- Update primary row
		PRINT 'D08 - Update primary NetworkDevice, DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		UPDATE	NetworkDevice
		SET	DuplicateID = NULLIF(@PrimaryDeviceID, DeviceID),
			ComputerID = @ComputerID,
			DeviceName = @DeviceName,
			DNSFullName = @DNSFullName,
			NTDomainName = @NTDomainName,
			IPAddress = @IPAddress,
			IPSubnet = @IPSubnet,
			IPSubnetMask = @IPSubnetMask,
			PhysicalAddress = @PhysicalAddress,
			DeviceRole = @DeviceRole,
			OperatingSystem = @OperatingSystem,
			OperatingSystemID = @OperatingSystemID,
			IsManaged = @IsManaged,
			Description = @Description,
			SystemDescription = @SystemDescription,
			SystemLocation = @SystemLocation,
			SystemContact = @SystemContact,
			FirstDiscovered = @FirstDiscovered,
			LastUpdate = GETDATE(),
			LastDataSourceName = @LastDataSourceName,
			LastDataSourceType = @LastDataSourceType,

			OpenPortsTCP = @OpenPortsTCP,
			OpenPortsUDP = @OpenPortsUDP,
			ScannedOperatingSystem = @ScannedOperatingSystem,

			ScannedOsType	= @ScannedOsType,
			ScannedOsVendor	= @ScannedOsVendor,
			ScannedOsFamily	= @ScannedOsFamily,
			ScannedOsGen	= @ScannedOsGen,
			ScannedMacAddress	= @ScannedMacAddress,
			ScannedMacVendor	= @ScannedMacVendor,
			
			SQLDiscoveredBy = CASE WHEN @IsSQL = 0 THEN '' ELSE
								CASE WHEN ISNULL(@IsSQL, -1) = -1 THEN SQLDiscoveredBy ELSE @SQLDiscoveredBy END
							END,
			SQLPorts = CASE WHEN @IsSQL = 0 THEN '' ELSE
								CASE WHEN ISNULL(@IsSQL, -1) = -1 THEN SQLPorts ELSE @SQLPorts END
							END,
			DeviceRoleOverride = @DeviceRoleOverride,
			NoOidMatch = @NoOidMatch
		WHERE DeviceID = @FirstDeviceID
		
		IF (@IsSQL <> -1)
		BEGIN
			exec NetworkDeviceCalculateMembers @FirstDeviceID
		END

		PRINT 'D10 - Update SNMP Information for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		-- Delete any SNMP info for this device and then repopulate
		-- if necessary
		PRINT 'D10 - Delete first the NetworkDeviceSNMPInfo for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		DELETE FROM NetworkDeviceSNMPInfo
			WHERE DeviceID = @FirstDeviceID

		IF (
			@SNMP_snmpCommunityString IS NOT NULL OR
			@SNMP_sysDescr IS NOT NULL OR
			@SNMP_sysObjectID IS NOT NULL OR
			@SNMP_sysObjectIDSymbolic IS NOT NULL OR
			@SNMP_sysUpTime IS NOT NULL OR
			@SNMP_sysContact IS NOT NULL OR
			@SNMP_sysName IS NOT NULL OR
			@SNMP_sysLocation IS NOT NULL OR
			@SNMP_sysServices IS NOT NULL OR
			@SNMP_ipForwarding IS NOT NULL
		)
		BEGIN
			PRINT 'D10 - Insert SNMPInformation to NetworkDeviceSNMPInfo table for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
			INSERT INTO NetworkDeviceSNMPInfo
			(
				DeviceID,
				SNMP_snmpCommunityString,
				SNMP_sysDescr,
				SNMP_sysObjectID,
				SNMP_sysObjectIDSymbolic,
				SNMP_sysUpTime,
				SNMP_sysContact,
				SNMP_sysName,
				SNMP_sysLocation,
				SNMP_sysServices,
				SNMP_ipForwarding
			)
			VALUES (
				@FirstDeviceID,
				@SNMP_snmpCommunityString,
				@SNMP_sysDescr,
				@SNMP_sysObjectID,
				@SNMP_sysObjectIDSymbolic,
				@SNMP_sysUpTime,
				@SNMP_sysContact,
				@SNMP_sysName,
				@SNMP_sysLocation,
				@SNMP_sysServices,
				@SNMP_ipForwarding
			)
		END
		ELSE
			PRINT 'D10 - no SNMP Information exists'

		PRINT 'D11 - Update AMT Information for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		-- Delete any AMT info for this device and then repopulate
		-- if necessary
		PRINT 'D11 - Delete first the NetworkDeviceAMTInfo for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		DELETE FROM NetworkDeviceAMTInfo
		WHERE DeviceID = @FirstDeviceID

		IF (
			@AMT_Secure IS NOT NULL OR
			@AMT_Description IS NOT NULL
		)
		BEGIN
			PRINT 'D11 - Insert AMT Information to NetworkDeviceAMTInfo table for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
			INSERT INTO NetworkDeviceAMTInfo
			(
				DeviceID,
				Secure,
				Description
			)
			VALUES (
				@FirstDeviceID,
				@AMT_Secure,
				@AMT_Description
			)
		END
		ELSE
			PRINT 'D11 - no AMT Information exists'

		PRINT 'D12 - Update Virtualization Information for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		-- Delete any Virtualization info for this device and then repopulate
		-- if necessary
		PRINT 'D12 - Delete first the NetworkDeviceVirtualisationInfo for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
		DELETE FROM NetworkDeviceVirtualizationInfo
		WHERE DeviceID = @FirstDeviceID

		IF (
			@Virtualization_Protocol IS NOT NULL OR
			@Virtualization_Port IS NOT NULL OR
			@Virtualization_APIType IS NOT NULL OR
			@Virtualization_APIVersion IS NOT NULL OR
			@Virtualization_NameID IS NOT NULL OR
			@Virtualization_Version IS NOT NULL
		)
		BEGIN
			PRINT 'D12 - Insert Virtualization Information to NetworkDeviceVirtualiszationInfo table for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
			INSERT INTO NetworkDeviceVirtualizationInfo
			(
				DeviceID,
				Protocol,
				Port,
				APIType,
				APIVersion,
				ProductNameID,
				ProductVersion
			)
			VALUES (
				@FirstDeviceID,
				@Virtualization_Protocol,
				@Virtualization_Port,
				@Virtualization_APIType,
				@Virtualization_APIVersion,
				@Virtualization_NameID,
				@Virtualization_Version
			)
		END
		ELSE
			PRINT 'D12 - no Virtualization Information exists'

	END
	ELSE
	BEGIN
		PRINT 'D13 - Create new NetworkDevice'
		PRINT 'D13 - ComputerID = ' + ISNULL(CAST(@ComputerID AS NVARCHAR(16)), '<null>') +
				' / DeviceName = ' + ISNULL(@DeviceName, '<null>') +
				' / DNSFullName = ' + ISNULL(@DNSFullName, '<null>') +
				' / NTDomainName = ' + ISNULL(@NTDomainName, '<null>') +
				' / IPAddress = ' + ISNULL(@IPAddress, '<null>') +
				' / IPSubnet = ' + ISNULL(@IPSubnet, '<null>') +
				' / IPSubnetMask = ' + ISNULL(@IPSubnetMask, '<null>') +
				' / PhysicalAddress = ' + ISNULL(@PhysicalAddress, '<null>') +
				' / DeviceRole = ' + ISNULL(CAST(@DeviceRole AS NVARCHAR(16)), '<null>') +
				' / NoOidMatch = ' + ISNULL(CAST(@NoOidMatch AS NVARCHAR(16)), '<null>') +
				' / OperatingSystemID = ' + ISNULL(CAST(@OperatingSystemID AS NVARCHAR(16)), '<null>') +
				' / OperatingSystem = ' + ISNULL(@OperatingSystem, '<null>') +
				' / IsManaged = ' + ISNULL(CAST(@IsManaged AS NVARCHAR(16)), '<null>') +
				' / Description = ' + ISNULL(@Description, '<null>') +
				' / SystemDescription = ' + ISNULL(@SystemDescription, '<null>') +
				' / SystemLocation = ' + ISNULL(@SystemLocation, '<null>') +
				' / SystemContact = ' + ISNULL(@SystemContact, '<null>') +
				' / FirstDiscovered = ' + ISNULL(CAST(@FirstDiscovered AS NVARCHAR(32)), '<null>')+
				' / LastDataSourceName = ' + ISNULL(@LastDataSourceName, '<null>') +
				' / LastDataSourceType = ' + ISNULL(CAST(@LastDataSourceType AS NVARCHAR(16)), '<null>') +
				' / OpenPortsTCP = ' + ISNULL(@OpenPortsTCP, '<null>') +
				' / OpenPortsUDP = ' + ISNULL(@OpenPortsUDP, '<null>') +
				' / ScannedOperatingSystem = ' + ISNULL(@ScannedOperatingSystem, '<null>') +
				' / ScannedOSType = ' + ISNULL(@ScannedOSType, '<null>') +
				' / ScannedOSVendor = ' + ISNULL(@ScannedOSVendor, '<null>') +
				' / ScannedOSFamily = ' + ISNULL(@ScannedOSFamily, '<null>') +
				' / ScannedOSGen = ' + ISNULL(@ScannedOSGen, '<null>') +
				' / ScannedMacAddress = ' + ISNULL(@ScannedMacAddress, '<null>') +
				' / ScannedMacVendor = ' + ISNULL(@ScannedMacVendor, '<null>') +
				' / SNMP_snmpCommunityString = ' + ISNULL(@SNMP_snmpCommunityString, '<null>') +
				' / SNMP_sysDescr = ' + ISNULL(@SNMP_sysDescr, '<null>') +
				' / SNMP_sysObjectID = ' + ISNULL(@SNMP_sysObjectID, '<null>') +
				' / SNMP_sysObjectIDSymbolic = ' + ISNULL(@SNMP_sysObjectIDSymbolic, '<null>') +
				' / SNMP_sysUpTime = ' + ISNULL(CAST(@SNMP_sysUpTime AS NVARCHAR(32)), '<null>') +
				' / SNMP_sysContact = ' + ISNULL(@SNMP_sysContact, '<null>') +
				' / SNMP_sysName = ' + ISNULL(@SNMP_sysName, '<null>') +
				' / SNMP_sysLocation = ' + ISNULL(@SNMP_sysLocation, '<null>') +
				' / SNMP_sysServices = ' + ISNULL(CAST(@SNMP_sysServices AS NVARCHAR(16)), '<null>') +
				' / SNMP_ipForwarding = ' + ISNULL(CAST(@SNMP_ipForwarding AS NVARCHAR(16)), '<null>') +
				' / AMT_Secure = ' + ISNULL(CAST(@AMT_Secure AS NVARCHAR(16)), '<null>') +
				' / AMT_Description = ' + ISNULL(@AMT_Description, '<null>') +
				' / Virtualization_Protocol = ' + ISNULL(@Virtualization_Protocol, '<null>') +
				' / Virtualization_Port = ' + ISNULL(CAST(@Virtualization_Port AS NVARCHAR(16)), '<null>') +
				' / Virtualization_APIType = ' + ISNULL(@Virtualization_APIType, '<null>') +
				' / Virtualization_APIVersion = ' + ISNULL(@Virtualization_APIVersion, '<null>') +
				' / Virtualization_NameID = ' + ISNULL(CAST(@Virtualization_NameID AS NVARCHAR(16)), '<null>') +
				' / Virtualization_Version = ' + ISNULL(@Virtualization_Version, '<null>') +
				' / DeviceRoleOverride = ' + ISNULL(CAST(@DeviceRoleOverride AS NVARCHAR(16)), '<null>')
		-- This is a new network device: insert new row
		PRINT 'D14 - Create NetworkDevice Table entry'
		INSERT INTO NetworkDevice
		(
			DuplicateID, ComputerID, DeviceName, DNSFullName,
			NTDomainName, IPAddress, IPSubnet,
			IPSubnetMask, PhysicalAddress, DeviceRole,
			OperatingSystem, OperatingSystemID, IsManaged,
			Description, SystemDescription, SystemLocation,
			SystemContact, LastDataSourceName, LastDataSourceType,

			OpenPortsTCP,
			OpenPortsUDP,
			ScannedOperatingSystem,

			ScannedOsType,
			ScannedOsVendor,
			ScannedOsFamily,
			ScannedOsGen,
			ScannedMacAddress,
			ScannedMacVendor,
			SQLDiscoveredBy,
			SQLPorts,
			DeviceRoleOverride,
			NoOidMatch
		)
		VALUES (
			@PrimaryDeviceID, @ComputerID, @DeviceName, @DNSFullName,
			@NTDomainName, @IPAddress, @IPSubnet,
			@IPSubnetMask, @PhysicalAddress, @DeviceRole,
			@OperatingSystem, @OperatingSystemID, @IsManaged,
			@Description, @SystemDescription, @SystemLocation,
			@SystemContact, @LastDataSourceName, @LastDataSourceType,

			@OpenPortsTCP,
			@OpenPortsUDP,
			@ScannedOperatingSystem,

			@ScannedOsType,
			@ScannedOsVendor,
			@ScannedOsFamily,
			@ScannedOsGen,
			@ScannedMacAddress,
			@ScannedMacVendor,
			CASE WHEN ISNULL(@IsSQL, 0) = 0 THEN '' ELSE @SQLDiscoveredBy END,
			@SQLPorts,
			@DeviceRoleOverride,
			@NoOidMatch
		)

		SET @FirstDeviceID = @@IDENTITY
		SET @PrimaryDeviceID = @FirstDeviceID
		
		IF (@IsSQL <> -1)
		BEGIN
			exec NetworkDeviceCalculateMembers @FirstDeviceID
		END

		-- Populate SNMP information if we have any
		IF (
			@SNMP_snmpCommunityString IS NOT NULL OR
			@SNMP_sysDescr IS NOT NULL OR
			@SNMP_sysObjectID IS NOT NULL OR
			@SNMP_sysObjectIDSymbolic IS NOT NULL OR
			@SNMP_sysUpTime IS NOT NULL OR
			@SNMP_sysContact IS NOT NULL OR
			@SNMP_sysName IS NOT NULL OR
			@SNMP_sysLocation IS NOT NULL OR
			@SNMP_sysServices IS NOT NULL OR
			@SNMP_ipForwarding IS NOT NULL
		)
		BEGIN
			PRINT 'D15 - Create NetworkDeviceSNMPInfo Table entry'
			INSERT INTO NetworkDeviceSNMPInfo
			(
				DeviceID,
				SNMP_snmpCommunityString,
				SNMP_sysDescr,
				SNMP_sysObjectID,
				SNMP_sysObjectIDSymbolic,
				SNMP_sysUpTime,
				SNMP_sysContact,
				SNMP_sysName,
				SNMP_sysLocation,
				SNMP_sysServices,
				SNMP_ipForwarding
			)
			VALUES (
				@FirstDeviceID,
			@SNMP_snmpCommunityString,
			@SNMP_sysDescr,
			@SNMP_sysObjectID,
			@SNMP_sysObjectIDSymbolic,
			@SNMP_sysUpTime,
			@SNMP_sysContact,
			@SNMP_sysName,
			@SNMP_sysLocation,
			@SNMP_sysServices,
			@SNMP_ipForwarding
		)
		END
		ELSE
			PRINT 'D15 - no SNMP Information exist'

		-- Populate AMT information if we have any
		IF (
			@AMT_Secure IS NOT NULL OR
			@AMT_Description IS NOT NULL
		)
		BEGIN
			PRINT 'D16 - Create NetworkDeviceAMTInfo Table entry'
			INSERT INTO NetworkDeviceAMTInfo
			(
				DeviceID,
				Secure,
				Description
			)
			VALUES (
				@FirstDeviceID,
				@AMT_Secure,
				@AMT_Description
			)
		END
		ELSE
			PRINT 'D16 - no AMT Information exist'

		-- Populate Virtualization information if we have any
		IF (
			@Virtualization_Protocol IS NOT NULL OR
			@Virtualization_Port IS NOT NULL OR
			@Virtualization_APIType IS NOT NULL OR
			@Virtualization_APIVersion IS NOT NULL OR
			@Virtualization_NameID IS NOT NULL OR
			@Virtualization_Version IS NOT NULL
		)
		BEGIN
			PRINT 'D17 - Create NetworkDeviceVirtualizationInfo Table entry'
			INSERT INTO NetworkDeviceVirtualizationInfo
			(
				DeviceID,
				Protocol,
				Port,
				APIType,
				APIVersion,
				ProductNameID,
				ProductVersion
			)
			VALUES (
				@FirstDeviceID,
				@Virtualization_Protocol,
				@Virtualization_Port,
				@Virtualization_APIType,
				@Virtualization_APIVersion,
				@Virtualization_NameID,
				@Virtualization_Version
			)
		END
		ELSE
			PRINT 'D17 - no Virtualization Information exist'
	END

	-- Cleanup cursor
	CLOSE DuplicateCursor
	DEALLOCATE DuplicateCursor

	PRINT 'D18 - Check if better OperatingSystem information is in inventory information'
	-- Check whether we have a better OperatingSystem value from Inventory:
	SELECT	@InvOperatingSystemID = OperatingSystemID
		FROM	Computer
		WHERE	ComputerID = @ComputerID
	IF (@InvOperatingSystemID IS NOT NULL)
	BEGIN
		-- If InvOperatingSystemID is better/more refined, use it instead
		SET @IsBetter = CASE WHEN (@OperatingSystemID IS NULL
					OR  @OperatingSystemID <= 1)
				THEN 1	-- Anything's better than NULL/unknown/other
				ELSE 0
				END
		IF @IsBetter = 0	--  Old may be better:
		    PRINT 'D18 - Old Information should be good, but start exec @IsBetter = OperatingSystemCategoryTest @OperatingSystemID=' + ISNULL(CAST(@OperatingSystemID AS NVARCHAR(16)), '<null>') + ',@InvOperatingSystemID=' + ISNULL(CAST(@InvOperatingSystemID AS NVARCHAR(16)), '<null>')
			-- If OldOperatingSystem looks ok, is it really better?
			EXEC @IsBetter = OperatingSystemCategoryTest
						@OperatingSystemID,
						@InvOperatingSystemID
		IF @IsBetter <> 0	-- Old definitely *is* better
		BEGIN
		    PRINT 'D18 - Inventory Information is better'
			SET @OperatingSystemID = @InvOperatingSystemID
			SELECT @OperatingSystem = OperatingSystemName
				FROM OperatingSystem
				WHERE OperatingSystemID = @OperatingSystemID

			PRINT 'D18 - Update OperatingSystemID (' + ISNULL(CAST(@OperatingSystemID AS NVARCHAR(16)), '<null>') + ') and OperatingSystem (' + ISNULL(@OperatingSystem, '<null>') + ') in NetworkDevice Table for DeviceID = ' + CAST(@FirstDeviceID AS NVARCHAR(16))
			UPDATE	NetworkDevice
			SET	OperatingSystem = @OperatingSystem,
				OperatingSystemID = @OperatingSystemID
			WHERE DeviceID = @FirstDeviceID
		END
	END
	ELSE IF (@OperatingSystemID IS NOT NULL)
	BEGIN
		PRINT 'D18 - Update OperatingSystemID (' + CAST(@OperatingSystemID AS NVARCHAR(16)) + ') in Computer Table for ComputerID = ' + CAST(@ComputerID AS NVARCHAR(16))
		UPDATE	Computer
		SET OperatingSystemID = @OperatingSystemID
		WHERE ComputerID = @ComputerID
	END

	-- Synchronize NULL entries in primary device
	IF (@PrimaryDeviceID IS NOT NULL AND @PrimaryDeviceID <> @FirstDeviceID)
	BEGIN
		PRINT 'D19 - Update PrimaryDevice Information in NetworkDevice for PrimaryDeviceID = ' + CAST(@PrimaryDeviceID AS NVARCHAR(16))
		UPDATE	NetworkDevice
		SET	ComputerID = ISNULL(ComputerID, @ComputerID),
			DeviceRole = ISNULL(DeviceRole, @DeviceRole),
			OperatingSystem = ISNULL(OperatingSystem, @OperatingSystem),
			OperatingSystemID = ISNULL(OperatingSystemID, @OperatingSystemID),
			IsManaged = ISNULL(IsManaged, @IsManaged),
			Description = ISNULL(Description, @Description),
			SystemDescription = ISNULL(SystemDescription, @SystemDescription),
			SystemLocation = ISNULL(SystemLocation, @SystemLocation),
			SystemContact = ISNULL(SystemContact, @SystemContact),

			OpenPortsTCP = ISNULL(OpenPortsTCP, @OpenPortsTCP),
			OpenPortsUDP = ISNULL(OpenPortsUDP, @OpenPortsUDP),
			ScannedOperatingSystem = ISNULL(ScannedOperatingSystem, @ScannedOperatingSystem),

			ScannedOsType = ISNULL(ScannedOsType, @ScannedOsType),
			ScannedOsVendor = ISNULL(ScannedOsVendor, @ScannedOsVendor),
			ScannedOsFamily = ISNULL(ScannedOsFamily, @ScannedOsFamily),
			ScannedOsGen	 = ISNULL(ScannedOsGen	, @ScannedOsGen	),
			ScannedMacAddress = ISNULL(ScannedMacAddress, @ScannedMacAddress),
			ScannedMacVendor = ISNULL(ScannedMacVendor, @ScannedMacVendor),
			DeviceRoleOverride = ISNULL(DeviceRoleOverride, @DeviceRoleOverride),
			NoOidMatch = ISNULL(NoOidMatch, @NoOidMatch)
		WHERE DeviceID = @PrimaryDeviceID

		-- If no SNMP info for PrimaryDevice, but exists for FirstDevice then
		-- create row for PrimaryDevice, otherwise if exists for FirstDevice then
		-- update it, otherwise do nothing
		IF (
			NOT EXISTS(SELECT * FROM NetworkDeviceSNMPInfo WHERE DeviceID = @PrimaryDeviceID)
			AND EXISTS(SELECT * FROM NetworkDeviceSNMPInfo WHERE DeviceID = @FirstDeviceID)
		)
		BEGIN
			PRINT 'D19 - Insert SNMP Information for PrimarayDeviceID (' + CAST(@PrimaryDeviceID AS NVARCHAR(16)) + ') with the Information from FirstDeviceID (' + CAST(@FirstDeviceID AS NVARCHAR(16)) + ')'
			INSERT INTO NetworkDeviceSNMPInfo
			(
				DeviceID,
				SNMP_snmpCommunityString,
				SNMP_sysDescr,
				SNMP_sysObjectID,
				SNMP_sysObjectIDSymbolic,
				SNMP_sysUpTime,
				SNMP_sysContact,
				SNMP_sysName,
				SNMP_sysLocation,
				SNMP_sysServices,
				SNMP_ipForwarding
			)
			VALUES (
				@PrimaryDeviceID,
				@SNMP_snmpCommunityString,
				@SNMP_sysDescr,
				@SNMP_sysObjectID,
				@SNMP_sysObjectIDSymbolic,
				@SNMP_sysUpTime,
				@SNMP_sysContact,
				@SNMP_sysName,
				@SNMP_sysLocation,
				@SNMP_sysServices,
				@SNMP_ipForwarding
			)
		END
		ELSE IF (
			EXISTS(SELECT * FROM NetworkDeviceSNMPInfo WHERE DeviceID = @PrimaryDeviceID)
		)
		BEGIN
			PRINT 'D19 - Update SNMP Information for PrimarayDeviceID (' + CAST(@PrimaryDeviceID AS NVARCHAR(16)) + ') with moe Information if this is NULL in the Primary Device'
			UPDATE	NetworkDeviceSNMPInfo
			SET
			SNMP_snmpCommunityString = ISNULL(SNMP_snmpCommunityString, @SNMP_snmpCommunityString),
			SNMP_sysDescr = ISNULL(SNMP_sysDescr, @SNMP_sysDescr),
			SNMP_sysObjectID = ISNULL(SNMP_sysObjectID, @SNMP_sysObjectID),
			SNMP_sysObjectIDSymbolic = ISNULL(SNMP_sysObjectIDSymbolic, @SNMP_sysObjectIDSymbolic),
			SNMP_sysUpTime = ISNULL(SNMP_sysUpTime, @SNMP_sysUpTime),
			SNMP_sysContact = ISNULL(SNMP_sysContact, @SNMP_sysContact),
			SNMP_sysName = ISNULL(SNMP_sysName, @SNMP_sysName),
			SNMP_sysLocation = ISNULL(SNMP_sysLocation, @SNMP_sysLocation),
			SNMP_sysServices = ISNULL(SNMP_sysServices, @SNMP_sysServices),
			SNMP_ipForwarding = ISNULL(SNMP_ipForwarding, @SNMP_ipForwarding)
			WHERE DeviceID = @PrimaryDeviceID
		END

		-- If no AMT info for PrimaryDevice, but exists for FirstDevice then
		-- create row for PrimaryDevice, otherwise if exists for FirstDevice then
		-- update it, otherwise do nothing
		IF (
			NOT EXISTS(SELECT * FROM NetworkDeviceAMTInfo WHERE DeviceID = @PrimaryDeviceID)
			AND EXISTS(SELECT * FROM NetworkDeviceAMTInfo WHERE DeviceID = @FirstDeviceID)
		)
		BEGIN
			PRINT 'D19 - Insert AMT Information for PrimarayDeviceID (' + CAST(@PrimaryDeviceID AS NVARCHAR(16)) + ') with the Information from FirstDeviceID (' + CAST(@FirstDeviceID AS NVARCHAR(16)) + ')'
			INSERT INTO NetworkDeviceAMTInfo
			(
				DeviceID,
				Secure,
				Description
			)
			VALUES (
				@PrimaryDeviceID,
				@AMT_Secure,
				@AMT_Description
			)
		END
		ELSE IF (
			EXISTS(SELECT * FROM NetworkDeviceAMTInfo WHERE DeviceID = @PrimaryDeviceID)
		)
		BEGIN
			PRINT 'D19 - Update AMT Information for PrimarayDeviceID (' + CAST(@PrimaryDeviceID AS NVARCHAR(16)) + ') with moe Information if this is NULL in the Primary Device'
			UPDATE	NetworkDeviceAMTInfo
			SET
				Secure = ISNULL(Secure, @AMT_Secure),
				Description = ISNULL(Description, @AMT_Description)
		WHERE DeviceID = @PrimaryDeviceID
		END

		-- If no Virtualization info for PrimaryDevice, but exists for FirstDevice then
		-- create row for PrimaryDevice, otherwise if exists for FirstDevice then
		-- update it, otherwise do nothing
		IF (
			NOT EXISTS(SELECT * FROM NetworkDeviceVirtualizationInfo WHERE DeviceID = @PrimaryDeviceID)
			AND EXISTS(SELECT * FROM NetworkDeviceVirtualizationInfo WHERE DeviceID = @FirstDeviceID)
		)
		BEGIN
			PRINT 'D19 - Insert Virtualization Information for PrimarayDeviceID (' + CAST(@PrimaryDeviceID AS NVARCHAR(16)) + ') with the Information from FirstDeviceID (' + CAST(@FirstDeviceID AS NVARCHAR(16)) + ')'
			INSERT INTO NetworkDeviceVirtualizationInfo
			(
				DeviceID,
				Protocol,
				Port,
				APIType,
				APIVersion,
				ProductNameID,
				ProductVersion
			)
			VALUES (
				@PrimaryDeviceID,
				@Virtualization_Protocol,
				@Virtualization_Port,
				@Virtualization_APIType,
				@Virtualization_APIVersion,
				@Virtualization_NameID,
				@Virtualization_Version
			)
		END
		ELSE IF (
			EXISTS(SELECT * FROM NetworkDeviceVirtualizationInfo WHERE DeviceID = @PrimaryDeviceID)
		)
		BEGIN
			PRINT 'D19 - Update Virtualization Information for PrimarayDeviceID (' + CAST(@PrimaryDeviceID AS NVARCHAR(16)) + ') with moe Information if this is NULL in the Primary Device'
			UPDATE	NetworkDeviceVirtualizationInfo
			SET
				Protocol = ISNULL(Protocol, @Virtualization_Protocol),
				Port = ISNULL(Port, @Virtualization_Port),
				APIType = ISNULL(APIType, @Virtualization_APIType),
				APIVersion = ISNULL(APIVersion, @Virtualization_APIVersion),
				ProductNameID = ISNULL(ProductNameID, @Virtualization_NameID),
				ProductVersion = ISNULL(ProductVersion, @Virtualization_Version)
		WHERE DeviceID = @PrimaryDeviceID
		END
	END


	-- Now we need to create/update the NetworkDeviceDiscoveredBy
	-- entry for this device.
	IF (@DiscoveringServerUID IS NOT NULL)
	BEGIN
		PRINT 'D20 - create/update the NetworkDeviceDiscoverdBy'
		DECLARE @DiscoveredByLastUpate DATETIME
		SET @DiscoveredByLastUpate = GETDATE()

		-- if there already exists a record for this device we simply
		-- update it with new values
		if (EXISTS(SELECT * FROM NetworkDeviceDiscoveredBy
			WHERE ServerUID = @DiscoveringServerUID AND DeviceID = @FirstDeviceID))
		BEGIN
			-- leave CanAdminister unchanged
			IF (@DiscoveringServerCanAdminister = 2)
				SELECT @DiscoveringServerCanAdminister = CanAdminister
					FROM NetworkDeviceDiscoveredBy
					WHERE ServerUID = @DiscoveringServerUID AND DeviceID = @FirstDeviceID

			PRINT 'D20 - Start exec NetworkDeviceDiscoveryByUpdateByDeviceIDServerUID @FirstDeviceID=' + CAST(@FirstDeviceID AS NVARCHAR(16)) + 
					',@DiscoveringServerUID=' + ISNULL(CAST(@DiscoveringServerUID AS NVARCHAR(32)), '<null>') + 
					',@DiscoveringServerCanAdminister=' + ISNULL(CAST(@DiscoveringServerCanAdminister AS NVARCHAR(16)), '<null>') + 
					',@DiscoveredByLastUpate=' + ISNULL(CAST(@DiscoveredByLastUpate AS NVARCHAR(16)), '<null>') +
					',@AccountID=' + ISNULL(CAST(@AccountID AS NVARCHAR(16)), '<null>')
			EXEC NetworkDeviceDiscoveredByUpdateByDeviceIDServerUID
				@FirstDeviceID,
				@DiscoveringServerUID,
				@DiscoveringServerCanAdminister,
				@DiscoveredByLastUpate,
				@AccountID
		END
		ELSE
		BEGIN
			-- we need to create a new discovered by record
			IF (@DiscoveringServerCanAdminister = 2)
				SET @DiscoveringServerCanAdminister = 0

			PRINT 'D20 - Start exec NetworkDeviceDiscoveryByAdd @FirstDeviceID=' + CAST(@FirstDeviceID AS NVARCHAR(16)) + 
					',@DiscoveringServerUID=' + ISNULL(CAST(@DiscoveringServerUID AS NVARCHAR(32)), '<null>') + 
					',@DiscoveringServerCanAdminister=' + ISNULL(CAST(@DiscoveringServerCanAdminister AS NVARCHAR(16)), '<null>') + 
					',@DiscoveredByLastUpate=' + ISNULL(CAST(@DiscoveredByLastUpate AS NVARCHAR(32)), '<null>') +
					',@AccountID=' + ISNULL(CAST(@AccountID AS NVARCHAR(16)), '<null>')
			EXEC NetworkDeviceDiscoveredByAdd
				@FirstDeviceID,
				@DiscoveringServerUID,
				@DiscoveringServerCanAdminister,
				@DiscoveredByLastUpate,
				@AccountID
		END
	END

	IF (@PrimaryDeviceID IS NOT NULL)
	BEGIN
		SET @DeviceID_OUT = @PrimaryDeviceID
	END
	ELSE
	BEGIN
		SET @DeviceID_OUT = @FirstDeviceID
	END

	PRINT 'D21 - End AddNetworkDeviceSysNameFirst with DeviceID=' + CAST(@DeviceID_OUT  AS NVARCHAR(16))

	COMMIT TRANSACTION
END


GO